
class APIError(IOError):

    def __init__(self, service_name, source, original_error):
        self.original_error = original_error
        self.service_name = service_name
        self.source = source

class HTTPError(APIError):

    def __init__(self, service_name, source, original_error, response):
        super().__init__(service_name, source, original_error)
        self.internal_status_code = response.status_code
        self.response = response.data

# For 4xx class errors
class HTTPClientError(HTTPError):
    pass

# For 5xx class errors
class HTTPServerError(HTTPError):
    pass

# For 400 HTTP errors
class BadRequestError(HTTPClientError):
    pass

# For 404 HTTP errors
class NotFoundError(HTTPClientError):
    pass

# For 500 HTTP errors
class InternalServerError(HTTPServerError):
    pass

HTTP_STATUS_CODE_CLASS_ERRORS = {
    range(400, 500): HTTPClientError,
    range(500, 600): HTTPServerError,
}

HTTP_STATUS_CODE_ERRORS = {
    400: BadRequestError,
    404: NotFoundError,
    500: InternalServerError,
}




import requests
from requests.exceptions import RequestException
from unittest import TestCase
from unittest.mock import MagicMock

from hubble_shuttle.http import ShuttleResponse
from hubble_shuttle.exceptions import APIError, HTTPError

MOCK_SERVICE_NAME = "a-service-name"
MOCK_SOURCE = "some_function_name"

class ShuttleAPIExceptionsTest(TestCase):
    def test_create_from_request_exception(self):
        request_error = RequestException()
        api_error = APIError(MOCK_SERVICE_NAME, MOCK_SOURCE, request_error)

        self.assertEqual(
            api_error.original_error, request_error, "Sets original_error correctly"
        )
        self.assertEqual(
            api_error.service_name, MOCK_SERVICE_NAME, "Sets service_name correctly"
        )
        self.assertEqual(api_error.source, MOCK_SOURCE, "Sets source correctly")

    def test_create_from_http_error(self):
        error_response = MagicMock()
        response = ShuttleResponse({"some_key": "some_value"}, 410)

        http_error = requests.exceptions.HTTPError(response=error_response)

        api_error = HTTPError(MOCK_SERVICE_NAME, MOCK_SOURCE, http_error, response)

        self.assertEqual(
            api_error.internal_status_code, 410, "Sets status code correctly",
        )
        self.assertEqual(
            api_error.response,
            {"some_key": "some_value"},
            "Sets the response correctly",
        )




import re
import requests
from requests.exceptions import RequestException
from requests.exceptions import HTTPError as RequestsHTTPError
from urllib.parse import urljoin

from .exceptions import *

class RequestsShuttleTransport:

    def __init__(self, **kwargs):
        self.api_endpoint = kwargs["api_endpoint"]
        self.headers = kwargs["headers"]
        self.query = kwargs["query"]
        self.request_content_type = kwargs["request_content_type"]

        if "service_name" in kwargs:
            self.service_name = kwargs["service_name"]
        else:
            self.service_name = type(self).__name__

    def get(self, url, **kwargs):
        return self.__http_request("get", url, **kwargs)

    def post(self, url, **kwargs):
        return self.__http_request("post", url, **kwargs)

    def put(self, url, **kwargs):
        return self.__http_request("put", url, **kwargs)

    def patch(self, url, **kwargs):
        return self.__http_request("patch", url, **kwargs)

    def delete(self, url, **kwargs):
        return self.__http_request("delete", url, **kwargs)

    def __http_request(self, method, url, **kwargs):
        try:
            response = requests.request(
                method,
                self.__prepare_request_url(url),
                **self.__prepare_request_args(**kwargs),
            )

            self.__raise_for_status(url, response)

            return self.__parse_response(response)
        except RequestException as error:
            raise APIError(self.service_name, url, error)

    def __prepare_request_url(self, url):
        return urljoin(
            self.api_endpoint,
            url
        )

    def __prepare_request_args(self, **kwargs):
        request_args = {}

        request_headers = self.__prepare_request_headers(kwargs.get("headers", {}))
        if request_headers:
            request_args.update({"headers": request_headers})

        request_query = self.__prepare_request_query(kwargs.get("query", {}))
        if request_query:
            request_args.update({"params": request_query})

        if "data" in kwargs:
            content_type = kwargs.get("content_type", self.request_content_type)
            if content_type == "application/x-www-form-urlencoded":
                request_args.update({"data": kwargs["data"]})
            elif content_type == "application/json":
                request_args.update({"json": kwargs["data"]})
            else:
                raise ValueError("Unknown content type for request: {}".format(content_type))

        return request_args

    def __prepare_request_headers(self, headers):
        request_headers = {}
        if self.headers:
            request_headers.update(self.headers)
        if headers:
            request_headers.update(headers)
        return request_headers

    def __prepare_request_query(self, query):
        request_query = {}
        if self.query:
            request_query.update(self.query)
        if query:
            request_query.update(query)
        return request_query

    def __raise_for_status(self, url, response):
        try:
            response.raise_for_status()
        except RequestsHTTPError as error:
            error_class = self.__map_http_error_class(error)
            raise error_class(self.service_name, url, error, self.__parse_response(response))

    def __map_http_error_class(self, error):
        if error.response.status_code in HTTP_STATUS_CODE_ERRORS:
            return HTTP_STATUS_CODE_ERRORS[error.response.status_code]

        for status_range in HTTP_STATUS_CODE_CLASS_ERRORS:
            if error.response.status_code in status_range:
                return HTTP_STATUS_CODE_CLASS_ERRORS[status_range]

        return HTTPError

    def __parse_response(self, response):
        content_type = response.headers.get('Content-Type', '')
        if re.match(r"^application/json( ?;.+)?$", content_type):
            data = response.json()
        elif re.match(r"^text/plain( ?;.+)?$", content_type):
            data = response.text
        else:
            data = response.content

        return ShuttleResponse(data, response.status_code)

class ShuttleResponse:

    def __init__(self, data, status_code):
        self.data = data
        self.status_code = status_code





from .http import RequestsShuttleTransport

class ShuttleAPI:

    transport = RequestsShuttleTransport

    api_endpoint = None
    headers = {}
    query = {}
    request_content_type = "application/x-www-form-urlencoded"

    def __init__(self, **kwargs):
        if "api_endpoint" in kwargs:
            self.api_endpoint = kwargs["api_endpoint"]
        if "headers" in kwargs:
            self.headers = kwargs["headers"]
        if "query" in kwargs:
            self.query = kwargs["query"]
        if "request_content_type" in kwargs:
            self.request_content_type = kwargs["request_content_type"]

        self.http = self.transport(
            api_endpoint = self.api_endpoint,
            headers = self.headers,
            query = self.query,
            request_content_type = self.request_content_type,
            service_name = type(self).__name__,
        )

    def http_get(self, url, **kwargs):
        return self.http.get(url, **kwargs)

    def http_post(self, url, **kwargs):
        return self.http.post(url, **kwargs)

    def http_put(self, url, **kwargs):
        return self.http.put(url, **kwargs)

    def http_patch(self, url, **kwargs):
        return self.http.patch(url, **kwargs)

    def http_delete(self, url, **kwargs):
        return self.http.delete(url, **kwargs)




import requests
from requests.exceptions import RequestException
from unittest import TestCase
from unittest.mock import MagicMock, patch

import hubble_shuttle
from hubble_shuttle import ShuttleAPI


class ShuttleAPITestClient(ShuttleAPI):
    api_endpoint = "http://test_http_server/"


class ShuttleAPITest(TestCase):

    def test_get_request(self):
        response = ShuttleAPITestClient().http_get("/get")
        self.assertEqual('http://test_http_server/get', response.data['url'], "Parses the JSON response")
        self.assertEqual({}, response.data['args'], "Doesn't include any query params")

    def test_get_request_status_code(self):
        response = ShuttleAPITestClient().http_get("/status/200")
        self.assertEqual(200, response.status_code, "Returns the HTTP status code")
        response = ShuttleAPITestClient().http_get("/status/201")
        self.assertEqual(201, response.status_code, "Returns the HTTP status code")

    def test_get_request_parse_response_text(self):
        response = ShuttleAPITestClient().http_get("/robots.txt")
        self.assertEqual("User-agent: *\nDisallow: /deny\n", response.data, "Returns the text response as a string")

    def test_get_request_parse_response_json(self):
        response = ShuttleAPITestClient().http_get("/json")
        self.assertEqual({'slideshow': {'author': 'Yours Truly',
                'date': 'date of publication',
                'slides': [{'title': 'Wake up to WonderWidgets!', 'type': 'all'},
                           {'items': ['Why <em>WonderWidgets</em> are great',
                                      'Who <em>buys</em> WonderWidgets'],
                            'title': 'Overview',
                            'type': 'all'}],
                'title': 'Sample Slide Show'}}, response.data, "Returns the JSON response as parsed object")

    def test_get_request_parse_response_default(self):
        response = ShuttleAPITestClient().http_get("/base64/SGVsbG8=")
        self.assertEqual(b"Hello", response.data, "Returns the response as a binary string")

    def test_get_gzip_content_encoding(self):
        response = ShuttleAPITestClient().http_get("/gzip")
        self.assertEqual(response.data['gzipped'], True, "Decodes the gzipped response content")

    def test_get_request_with_class_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_get("/get")
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")

    def test_get_request_with_request_headers(self):
        client = ShuttleAPITestClient()
        response = client.http_get("/get", headers={"Foo": "Bar"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_get_request_with_class_and_request_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_get("/get", headers={"Bar": "Baz"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")
        self.assertEqual("Baz", response.data['headers']['Bar'], "Sends the request-level headers")

    def test_get_request_with_class_and_request_headers_conflict(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_get("/get", headers={"Foo": "Baz"})
        self.assertEqual("Baz", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_get_request_with_class_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_get("/get")
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")

    def test_get_request_with_request_query_param(self):
        client = ShuttleAPITestClient()
        response = client.http_get("/get", query={"foo": "bar"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the request-level parameters")

    def test_get_request_with_class_and_request_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_get("/get", query={"bar": "baz"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")
        self.assertEqual("baz", response.data['args']['bar'], "Sends the request-level parameters")

    def test_get_request_with_class_and_request_query_param_conflict(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_get("/get", query={"foo": "baz"})
        self.assertEqual("baz", response.data['args']['foo'], "Sends the request-level parameters")

    def test_get_generic_networking_error(self):
        client = ShuttleAPITestClient(api_endpoint='http://test_http_server:1234')
        with self.assertRaises(hubble_shuttle.exceptions.APIError) as cm:
            client.http_get("/get")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/get", cm.exception.source, "Sets the error source")

    def test_get_400_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.BadRequestError) as cm:
            ShuttleAPITestClient().http_get("/status/400")
        self.assertEqual(400, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/400", cm.exception.source, "Sets the error source")

    def test_get_404_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.NotFoundError) as cm:
            ShuttleAPITestClient().http_get("/status/404")
        self.assertEqual(404, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/404", cm.exception.source, "Sets the error source")

    def test_get_499_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPClientError) as cm:
            ShuttleAPITestClient().http_get("/status/499")
        self.assertEqual(499, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/499", cm.exception.source, "Sets the error source")

    def test_get_500_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.InternalServerError) as cm:
            ShuttleAPITestClient().http_get("/status/500")
        self.assertEqual(500, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/500", cm.exception.source, "Sets the error source")

    def test_get_599_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPServerError) as cm:
            ShuttleAPITestClient().http_get("/status/599")
        self.assertEqual(599, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/599", cm.exception.source, "Sets the error source")

    def test_post_request(self):
        response = ShuttleAPITestClient().http_post("/post")
        self.assertEqual('http://test_http_server/post', response.data['url'], "Parses the JSON response")
        self.assertEqual({}, response.data['args'], "Doesn't include any query params")

    def test_post_request_no_body(self):
        response = ShuttleAPITestClient().http_post("/post")
        self.assertEqual("", response.data['data'], "Doesn't send any content body")
        self.assertEqual({}, response.data['form'], "Doesn't send any form data")

    def test_post_request_with_data(self):
        response = ShuttleAPITestClient().http_post("/post", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['form'], "Sends the data in form format")
        self.assertEqual("application/x-www-form-urlencoded", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_post_request_with_data_form_urlencoded(self):
        response = ShuttleAPITestClient().http_post("/post", content_type="application/x-www-form-urlencoded", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['form'], "Sends the data in form format")
        self.assertEqual("application/x-www-form-urlencoded", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_post_request_with_data_json(self):
        response = ShuttleAPITestClient().http_post("/post", content_type="application/json", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['json'], "Sends the data in JSON format")
        self.assertEqual("application/json", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_post_request_with_data_unknown_content_type(self):
        with self.assertRaises(ValueError) as cm:
            ShuttleAPITestClient().http_post("/post", content_type="application/bad-content-type", data={"foo": "bar", "bar": "baz"})
        self.assertEqual(str(cm.exception), "Unknown content type for request: application/bad-content-type", "Indicates the invalid content type")

    def test_post_request_with_data_client_request_content_type(self):
        client = ShuttleAPITestClient(request_content_type="application/json")
        response = client.http_post("/post", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['json'], "Sends the data in JSON format")
        self.assertEqual("application/json", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_post_request_status_code(self):
        response = ShuttleAPITestClient().http_post("/status/200")
        self.assertEqual(200, response.status_code, "Returns the HTTP status code")
        response = ShuttleAPITestClient().http_post("/status/201")
        self.assertEqual(201, response.status_code, "Returns the HTTP status code")

    def test_post_request_with_class_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_post("/post")
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")

    def test_post_request_with_request_headers(self):
        client = ShuttleAPITestClient()
        response = client.http_post("/post", headers={"Foo": "Bar"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_post_request_with_class_and_request_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_post("/post", headers={"Bar": "Baz"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")
        self.assertEqual("Baz", response.data['headers']['Bar'], "Sends the request-level headers")

    def test_post_request_with_class_and_request_headers_conflict(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_post("/post", headers={"Foo": "Baz"})
        self.assertEqual("Baz", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_post_request_with_class_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_post("/post")
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")

    def test_post_request_with_request_query_param(self):
        client = ShuttleAPITestClient()
        response = client.http_post("/post", query={"foo": "bar"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the request-level parameters")

    def test_post_request_with_class_and_request_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_post("/post", query={"bar": "baz"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")
        self.assertEqual("baz", response.data['args']['bar'], "Sends the request-level parameters")

    def test_post_request_with_class_and_request_query_param_conflict(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_post("/post", query={"foo": "baz"})
        self.assertEqual("baz", response.data['args']['foo'], "Sends the request-level parameters")

    def test_post_generic_networking_error(self):
        client = ShuttleAPITestClient(api_endpoint='http://test_http_server:1234')
        with self.assertRaises(hubble_shuttle.exceptions.APIError) as cm:
            client.http_post("/post")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/post", cm.exception.source, "Sets the error source")

    def test_post_400_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.BadRequestError) as cm:
            ShuttleAPITestClient().http_post("/status/400")
        self.assertEqual(400, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/400", cm.exception.source, "Sets the error source")

    def test_post_404_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.NotFoundError) as cm:
            ShuttleAPITestClient().http_post("/status/404")
        self.assertEqual(404, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/404", cm.exception.source, "Sets the error source")

    def test_post_499_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPClientError) as cm:
            ShuttleAPITestClient().http_post("/status/499")
        self.assertEqual(499, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/499", cm.exception.source, "Sets the error source")

    def test_post_500_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.InternalServerError) as cm:
            ShuttleAPITestClient().http_post("/status/500")
        self.assertEqual(500, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/500", cm.exception.source, "Sets the error source")

    def test_post_599_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPServerError) as cm:
            ShuttleAPITestClient().http_post("/status/599")
        self.assertEqual(599, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/599", cm.exception.source, "Sets the error source")

    def test_put_request(self):
        response = ShuttleAPITestClient().http_put("/put")
        self.assertEqual('http://test_http_server/put', response.data['url'], "Parses the JSON response")
        self.assertEqual({}, response.data['args'], "Doesn't include any query params")

    def test_put_request_no_body(self):
        response = ShuttleAPITestClient().http_put("/put")
        self.assertEqual("", response.data['data'], "Doesn't send any content body")
        self.assertEqual({}, response.data['form'], "Doesn't send any form data")

    def test_put_request_with_data(self):
        response = ShuttleAPITestClient().http_put("/put", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['form'], "Sends the data in form format")
        self.assertEqual("application/x-www-form-urlencoded", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_put_request_with_data_form_urlencoded(self):
        response = ShuttleAPITestClient().http_put("/put", content_type="application/x-www-form-urlencoded", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['form'], "Sends the data in form format")
        self.assertEqual("application/x-www-form-urlencoded", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_put_request_with_data_json(self):
        response = ShuttleAPITestClient().http_put("/put", content_type="application/json", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['json'], "Sends the data in JSON format")
        self.assertEqual("application/json", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_put_request_with_data_unknown_content_type(self):
        with self.assertRaises(ValueError) as cm:
            ShuttleAPITestClient().http_put("/put", content_type="application/bad-content-type", data={"foo": "bar", "bar": "baz"})
        self.assertEqual(str(cm.exception), "Unknown content type for request: application/bad-content-type", "Indicates the invalid content type")

    def test_put_request_with_data_client_request_content_type(self):
        client = ShuttleAPITestClient(request_content_type="application/json")
        response = client.http_put("/put", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['json'], "Sends the data in JSON format")
        self.assertEqual("application/json", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_put_request_status_code(self):
        response = ShuttleAPITestClient().http_put("/status/200")
        self.assertEqual(200, response.status_code, "Returns the HTTP status code")
        response = ShuttleAPITestClient().http_put("/status/201")
        self.assertEqual(201, response.status_code, "Returns the HTTP status code")

    def test_put_request_with_class_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_put("/put")
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")

    def test_put_request_with_request_headers(self):
        client = ShuttleAPITestClient()
        response = client.http_put("/put", headers={"Foo": "Bar"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_put_request_with_class_and_request_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_put("/put", headers={"Bar": "Baz"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")
        self.assertEqual("Baz", response.data['headers']['Bar'], "Sends the request-level headers")

    def test_put_request_with_class_and_request_headers_conflict(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_put("/put", headers={"Foo": "Baz"})
        self.assertEqual("Baz", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_put_request_with_class_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_put("/put")
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")

    def test_put_request_with_request_query_param(self):
        client = ShuttleAPITestClient()
        response = client.http_put("/put", query={"foo": "bar"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the request-level parameters")

    def test_put_request_with_class_and_request_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_put("/put", query={"bar": "baz"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")
        self.assertEqual("baz", response.data['args']['bar'], "Sends the request-level parameters")

    def test_put_request_with_class_and_request_query_param_conflict(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_put("/put", query={"foo": "baz"})
        self.assertEqual("baz", response.data['args']['foo'], "Sends the request-level parameters")

    def test_put_generic_networking_error(self):
        client = ShuttleAPITestClient(api_endpoint='http://test_http_server:1234')
        with self.assertRaises(hubble_shuttle.exceptions.APIError) as cm:
            client.http_put("/put")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/put", cm.exception.source, "Sets the error source")

    def test_put_400_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.BadRequestError) as cm:
            ShuttleAPITestClient().http_put("/status/400")
        self.assertEqual(400, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/400", cm.exception.source, "Sets the error source")

    def test_put_404_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.NotFoundError) as cm:
            ShuttleAPITestClient().http_put("/status/404")
        self.assertEqual(404, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/404", cm.exception.source, "Sets the error source")

    def test_put_499_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPClientError) as cm:
            ShuttleAPITestClient().http_put("/status/499")
        self.assertEqual(499, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/499", cm.exception.source, "Sets the error source")

    def test_put_500_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.InternalServerError) as cm:
            ShuttleAPITestClient().http_put("/status/500")
        self.assertEqual(500, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/500", cm.exception.source, "Sets the error source")

    def test_put_599_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPServerError) as cm:
            ShuttleAPITestClient().http_put("/status/599")
        self.assertEqual(599, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/599", cm.exception.source, "Sets the error source")

    def test_patch_request(self):
        response = ShuttleAPITestClient().http_patch("/patch")
        self.assertEqual('http://test_http_server/patch', response.data['url'], "Parses the JSON response")
        self.assertEqual({}, response.data['args'], "Doesn't include any query params")

    def test_patch_request_no_body(self):
        response = ShuttleAPITestClient().http_patch("/patch")
        self.assertEqual("", response.data['data'], "Doesn't send any content body")
        self.assertEqual({}, response.data['form'], "Doesn't send any form data")

    def test_patch_request_with_data(self):
        response = ShuttleAPITestClient().http_patch("/patch", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['form'], "Sends the data in form format")
        self.assertEqual("application/x-www-form-urlencoded", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_patch_request_with_data_form_urlencoded(self):
        response = ShuttleAPITestClient().http_patch("/patch", content_type="application/x-www-form-urlencoded", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['form'], "Sends the data in form format")
        self.assertEqual("application/x-www-form-urlencoded", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_patch_request_with_data_json(self):
        response = ShuttleAPITestClient().http_patch("/patch", content_type="application/json", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['json'], "Sends the data in JSON format")
        self.assertEqual("application/json", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_patch_request_with_data_unknown_content_type(self):
        with self.assertRaises(ValueError) as cm:
            ShuttleAPITestClient().http_patch("/patch", content_type="application/bad-content-type", data={"foo": "bar", "bar": "baz"})
        self.assertEqual(str(cm.exception), "Unknown content type for request: application/bad-content-type", "Indicates the invalid content type")

    def test_patch_request_with_data_client_request_content_type(self):
        client = ShuttleAPITestClient(request_content_type="application/json")
        response = client.http_patch("/patch", data={"foo": "bar", "bar": "baz"})
        self.assertEqual({"foo": "bar", "bar": "baz"}, response.data['json'], "Sends the data in JSON format")
        self.assertEqual("application/json", response.data['headers']['Content-Type'], "Sets the appropriate content type header")

    def test_patch_request_status_code(self):
        response = ShuttleAPITestClient().http_patch("/status/200")
        self.assertEqual(200, response.status_code, "Returns the HTTP status code")
        response = ShuttleAPITestClient().http_patch("/status/201")
        self.assertEqual(201, response.status_code, "Returns the HTTP status code")

    def test_patch_request_with_class_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_patch("/patch")
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")

    def test_patch_request_with_request_headers(self):
        client = ShuttleAPITestClient()
        response = client.http_patch("/patch", headers={"Foo": "Bar"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_patch_request_with_class_and_request_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_patch("/patch", headers={"Bar": "Baz"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")
        self.assertEqual("Baz", response.data['headers']['Bar'], "Sends the request-level headers")

    def test_patch_request_with_class_and_request_headers_conflict(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_patch("/patch", headers={"Foo": "Baz"})
        self.assertEqual("Baz", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_patch_request_with_class_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_patch("/patch")
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")

    def test_patch_request_with_request_query_param(self):
        client = ShuttleAPITestClient()
        response = client.http_patch("/patch", query={"foo": "bar"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the request-level parameters")

    def test_patch_request_with_class_and_request_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_patch("/patch", query={"bar": "baz"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")
        self.assertEqual("baz", response.data['args']['bar'], "Sends the request-level parameters")

    def test_patch_request_with_class_and_request_query_param_conflict(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_patch("/patch", query={"foo": "baz"})
        self.assertEqual("baz", response.data['args']['foo'], "Sends the request-level parameters")

    def test_patch_generic_networking_error(self):
        client = ShuttleAPITestClient(api_endpoint='http://test_http_server:1234')
        with self.assertRaises(hubble_shuttle.exceptions.APIError) as cm:
            client.http_patch("/patch")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/patch", cm.exception.source, "Sets the error source")

    def test_patch_400_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.BadRequestError) as cm:
            ShuttleAPITestClient().http_patch("/status/400")
        self.assertEqual(400, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/400", cm.exception.source, "Sets the error source")

    def test_patch_404_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.NotFoundError) as cm:
            ShuttleAPITestClient().http_patch("/status/404")
        self.assertEqual(404, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/404", cm.exception.source, "Sets the error source")

    def test_patch_499_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPClientError) as cm:
            ShuttleAPITestClient().http_patch("/status/499")
        self.assertEqual(499, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/499", cm.exception.source, "Sets the error source")

    def test_patch_500_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.InternalServerError) as cm:
            ShuttleAPITestClient().http_patch("/status/500")
        self.assertEqual(500, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/500", cm.exception.source, "Sets the error source")

    def test_patch_599_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPServerError) as cm:
            ShuttleAPITestClient().http_patch("/status/599")
        self.assertEqual(599, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/599", cm.exception.source, "Sets the error source")

    def test_delete_request(self):
        response = ShuttleAPITestClient().http_delete("/delete")
        self.assertEqual('http://test_http_server/delete', response.data['url'], "Parses the JSON response")
        self.assertEqual({}, response.data['args'], "Doesn't include any query params")

    def test_delete_request_status_code(self):
        response = ShuttleAPITestClient().http_delete("/status/200")
        self.assertEqual(200, response.status_code, "Returns the HTTP status code")
        response = ShuttleAPITestClient().http_delete("/status/201")
        self.assertEqual(201, response.status_code, "Returns the HTTP status code")

    def test_delete_request_with_class_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_delete("/delete")
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")

    def test_delete_request_with_request_headers(self):
        client = ShuttleAPITestClient()
        response = client.http_delete("/delete", headers={"Foo": "Bar"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_delete_request_with_class_and_request_headers(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_delete("/delete", headers={"Bar": "Baz"})
        self.assertEqual("Bar", response.data['headers']['Foo'], "Sends the client-level headers")
        self.assertEqual("Baz", response.data['headers']['Bar'], "Sends the request-level headers")

    def test_delete_request_with_class_and_request_headers_conflict(self):
        client = ShuttleAPITestClient(headers={"Foo": "Bar"})
        response = client.http_delete("/delete", headers={"Foo": "Baz"})
        self.assertEqual("Baz", response.data['headers']['Foo'], "Sends the request-level headers")

    def test_delete_request_with_class_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_delete("/delete")
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")

    def test_delete_request_with_request_query_param(self):
        client = ShuttleAPITestClient()
        response = client.http_delete("/delete", query={"foo": "bar"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the request-level parameters")

    def test_delete_request_with_class_and_request_query_param(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_delete("/delete", query={"bar": "baz"})
        self.assertEqual("bar", response.data['args']['foo'], "Sends the client-level parameters")
        self.assertEqual("baz", response.data['args']['bar'], "Sends the request-level parameters")

    def test_delete_request_with_class_and_request_query_param_conflict(self):
        client = ShuttleAPITestClient(query={"foo": "bar"})
        response = client.http_delete("/delete", query={"foo": "baz"})
        self.assertEqual("baz", response.data['args']['foo'], "Sends the request-level parameters")

    def test_delete_generic_networking_error(self):
        client = ShuttleAPITestClient(api_endpoint='http://test_http_server:1234')
        with self.assertRaises(hubble_shuttle.exceptions.APIError) as cm:
            client.http_delete("/delete")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/delete", cm.exception.source, "Sets the error source")

    def test_delete_400_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.BadRequestError) as cm:
            ShuttleAPITestClient().http_delete("/status/400")
        self.assertEqual(400, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/400", cm.exception.source, "Sets the error source")

    def test_delete_404_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.NotFoundError) as cm:
            ShuttleAPITestClient().http_delete("/status/404")
        self.assertEqual(404, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/404", cm.exception.source, "Sets the error source")

    def test_delete_499_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPClientError) as cm:
            ShuttleAPITestClient().http_delete("/status/499")
        self.assertEqual(499, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/499", cm.exception.source, "Sets the error source")

    def test_delete_500_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.InternalServerError) as cm:
            ShuttleAPITestClient().http_delete("/status/500")
        self.assertEqual(500, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/500", cm.exception.source, "Sets the error source")

    def test_delete_599_http_error(self):
        with self.assertRaises(hubble_shuttle.exceptions.HTTPServerError) as cm:
            ShuttleAPITestClient().http_delete("/status/599")
        self.assertEqual(599, cm.exception.internal_status_code, "Returns the error status code")
        self.assertEqual("ShuttleAPITestClient", cm.exception.service_name, "Sets the service name")
        self.assertEqual("/status/599", cm.exception.source, "Sets the error source")
